/*
 * TVision help to .info source.
 *
 * fro: Borland helpfile viewer for WWW v0.06 german patched
 *
 * search engine: Tomi Leppikangas <tomilepp@paju.oulu.fi>
 * www interface: Erik Bos <erik@xs4all.nl>
 * Adapted to generate a text output by: Molnar Lazlo <molnarl@postabank.hu>
 * Adapted to generate .info source by: Salvador E. Tropea <salvador@inti.edu.ar>
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>

#define EXAMPLE 	1
#define LINK		2
#define NOT_FOUND	-1

static char conv_table[14], *fname;
static int n_words;
static FILE *out;
static unsigned char *links;

typedef struct
{
   char *keyword;
   unsigned short link;
}
dir_type;

dir_type *directory;

static void die (char *text)
{
 printf ("AAAARGH!\nError: %s\n", text);
 exit (1);
}


inline static long OffSetOfLink(int incount)
{
 return links[incount*3]+links[incount*3+1]*256L+links[incount*3+2]*65536L;
}

char NameForNode[20];
static char *RealNameOf(int n,long pos)
{
 int i;
 for (i=0; i<n_words; i++)
    {
     if (directory[i].link==n)
        return directory[i].keyword;
    }
 sprintf(NameForNode,"Node_%ld", pos==-1 ? OffSetOfLink(n) : pos);
 return NameForNode;
}


static int IsText(char *s)
{
 if (isalpha(*s))
    return 1;
 if (*s==LINK)
   {
    for (s++;*s && *s!=LINK; s++);
    for (s++;*s && *s==' '; s++);
    return *s!='\n';
   }
 return 0;
}

static void print_as_info(char *s, unsigned short *links)
{
   int link_count = 0, example = 0, link = 0, first = 1;
   int in_format=1, in_col=0, in_table=0, columns=0, column=0;
   int column_w=0, i;

   while (*s)
   {
      switch (*s)
      {
//      case 0xfe:		/* list item */
//	 fprintf(out,"*");
//	 break;
#if 0
	 /* german Umlaute necessary in mosaic version 2.2,
	  * higher version ones (i don't know due to a lack of no
	  * other version handy yet)
	  *				J"org Abke
	  */
      case 0x94:
	 printf ("&ouml;");	/* "o */
	 break;
      case 0x99:
	 printf ("&Ouml;");	/* "O */
	 break;
      case 0x81:
	 printf ("&uuml;");	/* "u */
	 break;
      case 0x9a:
	 printf ("&Uuml;");	/* "U */
	 break;
      case 0x84:
	 printf ("&auml;");	/* "a */
	 break;
      case 0x8e:
	 printf ("&Auml;");	/* "A */
	 break;
      case 0xe1:		/*  "s */
	 printf ("&szlig;");
	 break;
#endif

      case EXAMPLE:
	 example ^= 1;
	 if (example)
	    fprintf(out,"@example\n");
	 else
	    fprintf(out,"@end example\n");
	 break;

      case '{':
           fprintf(out,"@{");
           break;

      case '}':
           fprintf(out,"@}");
           break;

      case '@':
           fprintf(out,"@@");
           break;

      case '\n':
           if (in_table)
             {
              if (*(s+1)!=0)
                {
                 if (columns==1)
                    fputc('',out);
                 fprintf(out,"\n ");
                }
              else
                {
                 fprintf(out,"\n ");
                 for (i=0; i<column_w; i++)
                     fputc('',out);
                 fprintf(out,"\n");
                }
              break;
             }
           if (example)
             {
              fputc(*s,out);
              break;
             }
           if (IsText(s+1) || (*(s+1)=='\n' && IsText(s+3)))
             {
              if (in_format)
                {
                 in_format=0;
                 fputc(*s,out);
                 fprintf(out,"@*\n@end format\n");
                }
              else
                fputc(*s,out);
             }
           else
             {
              if (!in_format)
                {
                 in_format=1;
                 fputc(*s,out);
                 fprintf(out,"@format\n");
                }
              else
                fputc(*s,out);
             }
           break;

      case '.':
           if (!in_format)
              fprintf(out,"@.");
           else
              fputc(*s,out);
           break;

//   That's the right way but I can control the spaces and other things
// I tried by I give up, so I used ...
/*      case LINK:
	 link ^= 1; // Flip-Flop
         if (link)
           {
            char *s=RealNameOf(links[link_count++],-1);
            fprintf(out,"@xref{%s, ,",s);
           }
         else
            fprintf(out,"}.");
	 break;*/
// brut force:
      case LINK:
	 link ^= 1; // Flip-Flop
         if (link)
           {
            fprintf(out,"*Note ");
            in_col=0;
           }
         else
           {
            char *stemp=RealNameOf(links[link_count++],-1);
            fprintf(out,": %s.",stemp);
            if (in_col && in_table)
              {
               fprintf(out," ");
               if (*(s+2)!=LINK)
                 { // End of table, draw the bottom
                  in_table=0;
                  column_w--;
                  while (++column!=columns)
                    {
                     for (i=0; i<column_w; i++)
                         fputc(' ',out);
                     fputc('',out);
                    }
                 fprintf(out,"\n ");
                 column=0;
                 while (column!=columns)
                   {
                    for (i=0; i<column_w; i++)
                        fputc('',out);
                    column++;
                    fputc(column!=columns ? '' : '',out);
                   }
                 fprintf(out,"\n");
                }
              else
                {
                 column++;
                 if (column==columns)
                    column=0; // Next line in the table
                 else
                    s++; // Skip the \n
                }
              }
           }
	 break;

      case '':
           if (link)
              in_col=1;
           else
              fputc(*s,out);
           break;

      case '':
           if (*(s-1)!=' ')
             {
              fputc(*s,out);
              break;
             }
           { // Skip frames, is too complex
            char *stemp=s;
            columns=1;
            while (*stemp!='\n' && *s)
              {
               if (*stemp=='')
                  columns++;
               stemp++;
              }
            if (columns==1)
               break;
           }
           in_table=1;
           fprintf(out,"");
           columns=1;
           column_w=0;
           while (*s!='\n' && *s)
             {
              if (*s=='')
                {
                 columns++;
                 column_w=0;
                }
              else
                column_w++;
              fputc(*s,out);
              s++;
             }
           if (*(s-1)!='')
              fputc('',out);
           fprintf(out,"\n ");
           break;

      case ' ':
           if (link)
              fprintf(out,"@ ");
           else
              fputc(*s,out);
           break;

/*      case ':':
           fprintf(out,"@:");
           break;

      case ',':
           fprintf(out,"@,");
           break;*/

      default:
	 fputc(*s,out);
         first=0;
      }
      s++;
   }
 if (example)
    fprintf(out,"@end example\n");
 if (!in_format)
    fprintf(out,"@format\n");
}

static unsigned read_short (int in)
{
    unsigned char temp[2];
    read (in,temp,2);
    return  temp[0]+temp[1]*256;
}

static void check_head (int in)
{
 char head[5];
 read (in, (void *) head, 5);
 if (memcmp ((void *) head, "TURBO", 5))
    die("Wrong filetype (TURBO id missing)\n");
}

static void find_eof (int in)
{
 char c;
 do
    read (in, (void *) &c, 1);
 while (c != 0x1a);
}

static void read_conv_table (int in)
{
 int i;

 lseek (in, 0x1f, SEEK_CUR);	/* Goto conv_table  */
 read (in, conv_table, 14);
 for (i = 0; i < 14; i++)	/* Change nil to newline */
    if (conv_table[i] == '\0')
 conv_table[i] = '\n';
}


static void decomp_dir (unsigned char array[])
{
   int a, len, copy_bytes;
   int current, unpack, here;
   unsigned short size;
   unsigned char byte;

   size = *(unsigned short*) array;	/* First two bytes are size */
   n_words = size;
   array += 2;
   /* Alloc memory for directory */
   directory = calloc (size, sizeof (dir_type));

   current = unpack = here = 0;
   for (a = 0; a < size; a++)
   {
      byte = array[current++];
      len = byte & 0x1f;	/* Lower 5 bits are lenght */
      copy_bytes = byte / 32;	/* Upper 3 bits are copy len */

      if ((directory[here].keyword = malloc (len + copy_bytes + 1)) == NULL)
	 die ("Malloc error\n");
      if (copy_bytes)		/* Copy if needed */
	 memcpy (directory[here].keyword,
		 directory[here - 1].keyword, copy_bytes);
      unpack = copy_bytes;
      while (len--)
	 directory[here].keyword[unpack++] = array[current++];
      directory[here].keyword[unpack] = '\0';
      directory[here].link = *((unsigned short *) &array[current]);
      here++;
      current += 2;		/* skip link */
   }

}

#define NODE ((which^=1) ? (*current) & 0x0f : (*current++) >>4)

static void uncompress_page (unsigned char *array, unsigned char *links)
{
   unsigned char n, c, *current, *page, *point;
   int which = 0, end = 1, counter = 0, link_count = 0;

   current = array;
   point = page = malloc (50000);

   while (end)
   {
      if ((n = NODE) == 0x0f)	/* Is packed */
      {
	 c = NODE;
	 c += NODE << 4;
	 switch (c)
	 {
	 case 0x01:		/* End mark */
	    end = 0;
	    *point = 0;
	    break;
	 case 0x05:		/* Example */
	    *point++ = EXAMPLE;
	    break;
	 case 0x02:		/* Link */
	    *point++ = LINK;
	    link_count++;	/* counted twise for every link */
	    break;
	 default:		/* Store */
	    do
	    {
	       *point++ = c;
	    }
	    while (counter--);
	    counter = 0;
	 }
      }
      else if (n == 0x0e)	/* Is counter */
	 counter = NODE + 1;
      else
      {
	 do
	 {
	    *point++ = conv_table[n];
	 }
	 while (counter--);
	 counter = 0;
      }
   }
   /* Check that we have right link table */
   if (links==NULL)
       return;
   links += 4;			/* Skip 4 zeros */
//   if ((*(unsigned short *) links) != link_count / 2)
//      printf ("Link count doesn't match!! %d %d\n",*(unsigned short *) links,link_count / 2);
   links += 2;
   print_as_info(page, (short *) links);
   free(page);
}

static void FilterInvalidIn(char *s)
{
 for (; *s; s++)
    {
     if (*s==',' || *s=='{' || *s=='}' || *s=='@')
        *s='_';

    }
}

void PrintHelp(void)
{
 printf("
  This program can convert the Turbo Vision help file into a .txi file
suitable to be converted to .info by makeinfo.
  The program works perfectly on my system converting the TVision help that
comes with BC++ 3.1, using the makeinfo that belongs to the texinfo
package 3.9 and using the switch --no-validate. This switch is needed because
I didn't make any effort to write Up, Next, etc labels in the @node entries.
  The program can convert other files but the convertion can be incomplete.
I tried with the BC++ 3.1 help and works but the generate .info isn't very
good.
  To browse the generated .info is better if you have an info viewer that
cuts the *Note labels and hides the real name of the links.

  Parameters:
  -f infilename
     Name of the TVision help file.

  -o outfilename
     Name of the output file.

  -i info-file
     Name of the output file generated by makeinfo.

SET
");
 exit(1);
}

int main (int argc, char **argv)
{
 int in, incount,i;
 unsigned short size, dir_size, links_size, links_count;
 unsigned char test, *array, *llinks, *larray;
 char curLet, *fout, *NodeName, *finfo;
 long page;

 printf("TVision help file to .info converter v1.0, by SET.\n"
        "Copyright 1997 by Salvador E. Tropea <salvador@inti.edu.ar>\n\n"
        "Parts of this program belongs to:\n"
        "Tomi Leppikangas <tomilepp@paju.oulu.fi>\n"
        "Erik Bos <erik@xs4all.nl>\n"
        "Molnar Lazlo <molnarl@postabank.hu>\n\n");

 if (argc==1)
   {
    printf("Use: %s -f infilename -o outfilename -i info-file [-h]\n",argv[0]);
    return 1;
   }

 fname=NULL;
 fout=NULL;
 finfo=NULL;
 for (i=1; i<argc; i++)
    {
     if (argv[i][0]=='-')
       {
        switch (argv[i][1])
          {
           case 'f':
                if (argc>i+1)
                  {
                   i++;
                   fname=argv[i];
                  }
                break;

           case 'o':
                if (argc>i+1)
                  {
                   i++;
                   fout=argv[i];
                  }
                break;

           case 'i':
                if (argc>i+1)
                  {
                   i++;
                   finfo=argv[i];
                  }
                break;

           case 'h':
                PrintHelp();
                break;

           default:
                printf("Unknown command line option: %c\n",argv[i][1]);
                return 1;
          }
       }
     else
       printf("Unknown command line parameter: %s\n",argv[i]);
    }

 if (!fname)
   {
    puts("Missing input file name");
    return 1;
   }
 if (!fout)
   {
    puts("Missing output file name");
    return 1;
   }
 if (!finfo)
   {
    puts("Missing info name");
    return 1;
   }

 if ((in = open (fname, O_RDONLY|O_BINARY)) == -1)
   {
    printf("Can't open help file: %s.",fname);
    return 2;
   }

 if ((out=fopen(fout,"wt"))==NULL)
   {
    printf("Can't create the output file: %s.",fout);
    return 2;
   }

 // Check if starts with 'TURBO'
 check_head (in);
 // eat the rest of the name ' C HELP ...'
 find_eof (in);
 // Read the conv table, needed to unpack the node text
 read_conv_table (in);

 //  Read the table with the offsets of the links, each offset is a 3 bytes
 // pointer into the file (puaj :-[).
 lseek(in, 1, SEEK_CUR);	/* Skip one char */
 links_size=read_short (in);
 links_count=read_short (in);
 printf("Links count: %d, size: %d\n",links_count,links_size);
 if (!(links = malloc (links_size+100)))
    die ("Malloc error 1\n");
 read (in, links, links_size - 2);	/* 1 link = 3 bytes */

 //   Read the directory, is the index of the help, the data is packed.
 read (in, (void *) &test, 1);
 dir_size=read_short (in);
 if (!(larray = malloc (dir_size)))
    die ("Malloc error 2\n");
 read (in, larray, dir_size);
 decomp_dir (larray);


 fprintf(out,"@setfilename %s",finfo);
 fprintf(out,"

@direntry
* TVision 1.03: (tvision).           Documentation for TVision 1.03.
@end direntry

@titlepage
@end titlepage
");

 fprintf(out,"@node Top, a, (dir), (dir)\n");
 fprintf(out,"This is the documentation for TVision 1.03\n\n"
              "@menu\n");

 // Generate a page with the directory, n_words is the directory length
 curLet='@';
 printf("Generating the directory ");
 n_words=*((unsigned short *)larray);
 for (incount=0; incount<n_words; incount++)
    {
     if (toupper(directory[incount].keyword[0])>='A' &&
         toupper(directory[incount].keyword[0])<='Z')
        while (toupper(directory[incount].keyword[0])!=curLet)
          {
           fprintf(out,"\n %c\n\n",++curLet);
           printf(" %c",curLet);
           fflush(stdout);
          }
     // Cut invalid chars like ',' not supported in a node name
     FilterInvalidIn(directory[incount].keyword);
     fprintf(out,"* %s::.\n",directory[incount].keyword);
    }
 fprintf(out,"@end menu\n");
 printf("\n");

 // Now generate the nodes
 for (incount=0; incount<links_count; incount++)
    {
     printf("Generating node: %d\r",incount);
     fflush(stdout);
     // Find the position of the link
     page = OffSetOfLink(incount);

     // What's the purpose of this links? I don't know but exists
     if (page==0xFFFFFF)
        continue;

     // Create a name for it
     NodeName=RealNameOf(incount,page);
     fprintf(out,"@node %s\n@chapter %s\n@format\n",NodeName,NodeName);

     // Jump to it
     lseek (in, page + 1, SEEK_SET);	/* Move to page */
     // Read it
     size=read_short (in);             /* Read page */
     if (!(array = malloc (size)))
        die ("Malloc error 3\n");
     read (in, array, size);

     // Have crossreferences?
     read (in, (void *) &test, 1);
     if (test == 0x03)
       {
        size=read_short (in);          /* Read page-links */
        if (!(llinks = malloc (size)))
           die ("Malloc error 3\n");
        read (in, llinks, size);
       }
     else
        links=NULL;

     uncompress_page(array, llinks);

     fprintf(out,"@end format\n");

     free(array);
     free(llinks);
    }
 printf("\nSuccesfully converted! (no crashes ;-)\n");

 free(links);
 free(larray);
 close(in);
 fclose(out);

 return 0;
}
